# Generated by Django 4.2.11 on 2024-09-02 10:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('capeditor', '0008_remove_capsetting_audience_types'),
    ]

    operations = [
        migrations.AddField(
            model_name='capsetting',
            name='un_country_boundary_geojson',
            field=models.JSONField(blank=True, null=True, verbose_name='UN Country Boundary GeoJSON'),
        ),
    ]



# Generated by Django 4.2.7 on 2024-06-03 12:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('capeditor', '0007_alter_capsetting_wmo_oid_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='capsetting',
            name='audience_types',
        ),
    ]



import json

from django.contrib.gis.db import models
from django.utils.translation import gettext_lazy as _
from modelcluster.fields import ParentalKey
from modelcluster.models import ClusterableModel
from wagtail.admin.panels import FieldPanel, InlinePanel, TabbedInterface, ObjectList
from wagtail.contrib.settings.models import BaseSiteSetting
from wagtail.contrib.settings.registry import register_setting
from wagtail.fields import StreamField
from wagtail.models import Site, Orderable
from wagtailiconchooser.widgets import IconChooserWidget
from wagtailmodelchooser import register_model_chooser

from capeditor.blocks import (
    ContactBlock
)
from capeditor.forms.widgets import (
    HazardEventTypeWidget,
    MultiPolygonWidget,
    GeojsonFileLoaderWidget
)


@register_setting
class CapSetting(BaseSiteSetting, ClusterableModel):
    sender = models.CharField(max_length=255, blank=True, null=True, verbose_name=_("CAP Sender Email"),
                              help_text=_("Email of the sending institution"))
    sender_name = models.CharField(max_length=255, blank=True, null=True, verbose_name=_("CAP Sender Name"),
                                   help_text=_("Name of the sending institution"))
    wmo_oid = models.CharField(max_length=255, blank=True, null=True,
                               verbose_name=_("WMO Register of Alerting Authorities OID"),
                               help_text=_("WMO Register of Alerting Authorities "
                                           "Object Identifier (OID) of the sending institution. "
                                           "This will be used to generate CAP messages identifiers"))

    logo = models.ForeignKey("wagtailimages.Image", null=True, blank=True, on_delete=models.SET_NULL, related_name="+",
                             verbose_name=_("Logo of the sending institution"))

    contacts = StreamField([
        ("contact", ContactBlock(label=_("Contact")))
    ], use_json_field=True, blank=True, null=True, verbose_name=_("Contact Details"),
        help_text=_("Contact for follow-up and confirmation of the alert message"))

    un_country_boundary_geojson = models.JSONField(blank=True, null=True, verbose_name=_("UN Country Boundary"),
                                                   help_text=_("GeoJSON file of the UN Country Boundary. Setting this"
                                                               " will enable the UN Country Boundary check in the alert"
                                                               "drawing tools"))

    class Meta:
        verbose_name = _("CAP Settings")

    edit_handler = TabbedInterface([
        ObjectList([
            FieldPanel("sender_name"),
            FieldPanel("sender"),
            FieldPanel("wmo_oid"),
            FieldPanel("logo"),
            FieldPanel("contacts"),
        ], heading=_("Sender Details")),
        ObjectList([
            InlinePanel("hazard_event_types", heading=_("Hazard Types"), label=_("Hazard Type"),
                        help_text=_("Hazards monitored by the institution")),
        ], heading=_("Hazard Types")),
        ObjectList([
            InlinePanel("predefined_alert_areas", heading=_("Predefined Alert Areas"), label=_("Area"),
                        help_text=_("Predefined areas for alerts")),
        ], heading=_("Predefined Areas"), classname="map-resize-trigger"),
        ObjectList([
            FieldPanel("un_country_boundary_geojson",
                       widget=GeojsonFileLoaderWidget(
                           attrs={"resize_trigger_selector": ".w-tabs__tab.map-resize-trigger"})),
        ], heading=_("UN Boundary"), classname="map-resize-trigger"),
    ])

    @property
    def contact_list(self):
        contacts = []
        for contact_block in self.contacts:
            contact = contact_block.value.get("contact")
            if contact:
                contacts.append(contact)
        return contacts

    @property
    def audience_list(self):
        audiences = []
        for audience_block in self.audience_types:
            audience = audience_block.value.get("audience")
            if audience:
                audiences.append(audience)
        return audiences


class HazardEventTypes(Orderable):
    setting = ParentalKey(CapSetting, on_delete=models.PROTECT, related_name="hazard_event_types")
    is_in_wmo_event_types_list = models.BooleanField(default=True,
                                                     verbose_name=_("Select from WMO list of Hazards Event Types"))
    event = models.CharField(max_length=35, unique=True, verbose_name=_("Hazard"), help_text=_("Name of Hazard"))
    icon = models.CharField(max_length=255, null=True, blank=True, verbose_name=_("Icon"), help_text=_("Matching icon"))

    panels = [
        FieldPanel("is_in_wmo_event_types_list"),
        FieldPanel("event", widget=HazardEventTypeWidget),
        FieldPanel("icon", widget=IconChooserWidget),
    ]

    def __str__(self):
        return self.event


class PredefinedAlertArea(Orderable):
    setting = ParentalKey(CapSetting, on_delete=models.PROTECT, related_name="predefined_alert_areas")
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    geom = models.MultiPolygonField(srid=4326, verbose_name=_("Area"))

    class Meta:
        verbose_name = _("Predefined Area")
        verbose_name_plural = _("Predefined Areas")

    def __str__(self):
        return self.name

    @property
    def geojson(self):
        return json.loads(self.geom.geojson)

    panels = [
        FieldPanel("name"),
        FieldPanel("geom",
                   widget=MultiPolygonWidget(attrs={"resize_trigger_selector": ".w-tabs__tab.map-resize-trigger"})),
    ]


register_model_chooser(PredefinedAlertArea)


def get_cap_setting():
    try:
        site = Site.objects.get(is_default_site=True)
        if site:
            return CapSetting.for_site(site)
    except Exception:
        pass
    return None


def get_default_sender():
    cap_setting = get_cap_setting()
    if cap_setting and cap_setting.sender:
        return cap_setting.sender
    return None


def get_cap_contact_list(request):
    cap_settings = CapSetting.for_request(request)
    contacts_list = cap_settings.contact_list
    return contacts_list


def get_cap_audience_list(request):
    cap_settings = CapSetting.for_request(request)
    audience_list = cap_settings.audience_list
    return audience_list



import datetime
import xml.etree.ElementTree as ET
from io import StringIO

from rest_framework_xml.renderers import XMLRenderer


class CapXMLRenderer(XMLRenderer):
    format = 'xml'
    root_tag_name = 'alert'

    def _recursive_serialize_dict(self, value, parent):
        for key, value in value.items():
            if isinstance(value, list):
                # handle MULTIPOLYGONS
                if key == 'polygons':
                    for polygon_coords in value:
                        element = ET.SubElement(parent, "cap:polygon")
                        self._recursive_serialize(polygon_coords, element)
                else:
                    for item in value:
                        element = ET.SubElement(parent, f"cap:{key}")
                        self._recursive_serialize(item, element)

            else:
                element = ET.SubElement(parent, f"cap:{key}")
                self._recursive_serialize(value, element)

    def _recursive_serialize(self, value, element):
        if isinstance(value, dict):
            self._recursive_serialize_dict(value, element)
        elif isinstance(value, datetime.datetime):
            element.text = value.isoformat()
        elif isinstance(value, (int, float)):
            element.text = str(value)
        elif isinstance(value, str):
            element.text = value
        elif value is None:
            element.set('nil', 'true')
        else:
            element.text = str(value)

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Render `data` into XML.
        """
        tree = ET.ElementTree()
        stream = StringIO()

        if data is None:
            return ''

        if not isinstance(data, dict):
            raise ValueError('Data should be a dictionary')

        root = ET.Element('cap:alert', attrib={'xmlns:cap': 'urn:oasis:names:tc:emergency:cap:1.2'})

        self._recursive_serialize_dict(data, root)
        tree._setroot(root)
        tree.write(stream, encoding="unicode", xml_declaration=True)

        return stream.getvalue()



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import logging
import random

from paho.mqtt import client as mqtt_client

from .base import BasePubSubClient

LOGGER = logging.getLogger(__name__)


class MQTTPubSubClient(BasePubSubClient):
    """MQTT Pub/Sub client"""

    def __init__(self, broker: dict) -> None:
        """
        Pub/Sub initializer

        :param broker: `str` of broker RFC1738 URL

        :returns: `None`
        """

        super().__init__(broker)
        self.type = 'mqtt'
        self._port = self.broker_url.port
        self.client_id = f"cap-mqtt-{self.broker['client_type']}-{random.randint(0, 1000)}"  # noqa

        msg = f'Connecting to broker {self.broker} with id {self.client_id}'
        LOGGER.debug(msg)
        self.conn = mqtt_client.Client(self.client_id)

        self.conn.enable_logger(logger=LOGGER)

        if None not in [self.broker_url.password, self.broker_url.password]:
            self.conn.username_pw_set(
                self.broker_url.username,
                self.broker_url.password)

        if self._port is None:
            if self.broker_url.scheme == 'mqtts':
                self._port = 8883
            else:
                self._port = 1883
        if self.broker_url.scheme == 'mqtts':
            self.conn.tls_set(tls_version=2)

        self.conn.connect(self.broker_url.hostname, self._port)
        LOGGER.debug('Connected to broker')

    def pub(self, topic: str, message: str, qos: int = 1) -> bool:
        """
        Publish a message to a broker/topic

        :param topic: `str` of topic
        :param message: `str` of message

        :returns: `bool` of publish result
        """

        LOGGER.debug(f'Publishing to broker {self.broker}')
        LOGGER.debug(f'Topic: {topic}')
        LOGGER.debug(f'Message: {message}')

        result = self.conn.publish(topic, message, qos)

        # TODO: investigate implication
        # result.wait_for_publish()

        if result.is_published:
            return True
        else:
            msg = f'Publishing error code: {result[1]}'
            LOGGER.warning(msg)
            return False

    def sub(self, topic: str) -> None:
        """
        Subscribe to a broker/topic

        :param topic: `str` of topic

        :returns: `None`
        """

        def on_connect(client, userdata, flags, rc):
            LOGGER.debug(f'Connected to broker {self.broker}')
            LOGGER.debug(f'Subscribing to topic {topic} ')
            client.subscribe(topic, qos=1)
            LOGGER.debug(f'Subscribed to topic {topic}')

        def on_disconnect(client, userdata, rc):
            LOGGER.debug(f'Disconnected from {self.broker}')

        LOGGER.debug(f'Subscribing to broker {self.broker}, topic {topic}')
        self.conn.on_connect = on_connect
        self.conn.on_disconnect = on_disconnect
        self.conn.loop_forever()

    def __repr__(self):
        return '<MQTTPubSubClient>'



from datetime import datetime

import pytz
from shapely import Point
from xmltodict import parse as parse_xml_to_dict

from capeditor.errors import CAPImportError

CAP_ALERT_ELEMENTS = [
    {
        "name": "identifier",
        "required": True,
    },
    {
        "name": "sender",
        "required": True,
    },
    {
        "name": "sent",
        "required": True,
        "datetime": True,
    },
    {
        "name": "status",
        "required": True,
    },
    {
        "name": "msgType",
        "required": True,
    },
    {
        "name": "scope",
        "required": True,
    },
    {
        "name": "restriction",
        "required": False,
    },
    {
        "name": "addresses",
        "required": False,
    },
    {
        "name": "code",
        "required": False,
        "many": True,
    },
    {
        "name": "note",
        "required": False,
    },
    {
        "name": "references",
        "required": False,
    },
    {
        "name": "incidents",
        "required": False,
    },
    {
        "name": "info",
        "required": True,
        "many": True,
        "elements": [
            {
                "name": "language",
                "required": False,
            },
            {
                "name": "category",
                "required": True,
                "many": True,
            },
            {
                "name": "event",
                "required": True,
            },
            {
                "name": "responseType",
                "required": False,
                "many": True,
            },
            {
                "name": "urgency",
                "required": True,
            },
            {
                "name": "severity",
                "required": True,
            },
            {
                "name": "certainty",
                "required": True,
            },
            {
                "name": "audience",
                "required": False,
            },
            {
                "name": "eventCode",
                "required": False,
                "many": True,
                "elements": [
                    {
                        "name": "valueName",
                        "required": True,
                    },
                    {
                        "name": "value",
                        "required": True,
                    }
                ]
            },
            {
                "name": "effective",
                "required": False,
                "datetime": True,
            },
            {
                "name": "onset",
                "required": False,
                "datetime": True,
            },
            {
                "name": "expires",
                "required": False,
                "datetime": True,
            },
            {
                "name": "senderName",
                "required": False,
            },
            {
                "name": "headline",
                "required": True,
            },
            {
                "name": "description",
                "required": True,
            },
            {
                "name": "instruction",
                "required": False,
            },
            {
                "name": "web",
                "required": False,
            },
            {
                "name": "contact",
                "required": False,
            },
            {
                "name": "parameter",
                "required": False,
                "many": True,
                "elements": [
                    {
                        "name": "valueName",
                        "required": True,
                    },
                    {
                        "name": "value",
                        "required": True,
                    }
                ]
            },
            {
                "name": "resource",
                "required": False,
                "many": True,
                "elements": [
                    {
                        "name": "resourceDesc",
                        "required": True,
                    },
                    {
                        "name": "mimeType",
                        "required": True,
                    },
                    {
                        "name": "size",
                        "required": False,
                    },
                    {
                        "name": "uri",
                        "required": False,
                    },
                    {
                        "name": "derefUri",
                        "required": False,
                    },
                    {
                        "name": "digest",
                        "required": False,
                    }
                ]
            },
            {
                "name": "area",
                "required": True,
                "many": True,
                "geo": True,
                "elements": [
                    {
                        "name": "areaDesc",
                        "required": True,
                    },
                    {
                        "name": "polygon",
                        "required": False,
                        "many": True,
                    },
                    {
                        "name": "circle",
                        "required": False,
                        "many": True,
                    },
                    {
                        "name": "geocode",
                        "required": False,
                        "many": True,
                        "elements": [
                            {
                                "name": "valueName",
                                "required": True,
                            },
                            {
                                "name": "value",
                                "required": True,
                            }
                        ]
                    },
                    {
                        "name": "altitude",
                        "required": False,
                    },
                    {
                        "name": "ceiling",
                        "required": False,
                    }
                ]
            }
        ]
    }
]


def extract_element_data(element, data, validate=True):
    """
    Extract element data.

    :param element: The element to extract.
    :param data: The data to extract from.
    :param validate: Whether to validate the data for required fields.
    :return: The extracted element data.
    """

    element_name = element["name"]
    element_data = data.get(element_name) or data.get(f"cap:{element_name}")

    if validate and element["required"] and not element_data:
        raise CAPImportError(f"Missing required element: {element_name}")

    if not element_data:
        return None

    if element.get("datetime"):
        element_data = datetime.fromisoformat(element_data)
        element_data = element_data.astimezone(pytz.utc)
        element_data = element_data.isoformat()

    if element.get("many"):
        if not isinstance(element_data, list):
            element_data = [element_data]

    if element.get("elements"):
        if isinstance(element_data, list):
            for i, item in enumerate(element_data):
                element_data[i] = {}
                for sub_element in element["elements"]:
                    sub_element_data = extract_element_data(sub_element, item, validate=validate)
                    sub_element_name = sub_element["name"]

                    if sub_element_data:
                        element_data[i][sub_element_name] = sub_element_data

                    if sub_element_name == "polygon":
                        if sub_element_data:
                            polygons = []
                            for polygon in sub_element_data:
                                coordinates = []
                                for point in polygon.split(" "):
                                    lat, lon = point.split(",")
                                    coordinates.append([float(lon), float(lat)])
                                polygons.append([coordinates])

                            if len(polygons) > 1:
                                geometry = {
                                    "type": "MultiPolygon",
                                    "coordinates": polygons
                                }
                            else:
                                geometry = {
                                    "type": "Polygon",
                                    "coordinates": polygons[0]
                                }
                            element_data[i]["geometry"] = geometry

                    if sub_element_name == "circle":
                        if sub_element_data:
                            polygons = []
                            for circle in sub_element_data:
                                parts = circle.split()
                                coords = parts[0].split(',')
                                # Extract the longitude, latitude, and radius
                                longitude, latitude, radius_km = float(coords[0]), float(coords[1]), float(parts[1])
                                # Create a point for the center
                                center_point = Point(longitude, latitude)

                                # Convert radius to degrees (approximation for small distances)
                                radius_deg = radius_km / 111.12

                                circle = center_point.buffer(radius_deg)
                                coordinates = [[y, x] for x, y in list(circle.exterior.coords)]

                                polygons.append([coordinates])

                            if len(polygons) > 1:
                                geometry = {
                                    "type": "MultiPolygon",
                                    "coordinates": polygons
                                }
                            else:
                                geometry = {
                                    "type": "Polygon",
                                    "coordinates": polygons[0]
                                }
                            element_data[i]["geometry"] = geometry
        else:
            for sub_element in element["elements"]:
                sub_element_data = extract_element_data(sub_element, element_data, validate=validate)
                sub_element_name = sub_element["name"]

                if sub_element_data:
                    element_data[sub_element_name] = sub_element_data

    return element_data


def cap_xml_to_alert_data(cap_xml_string, validate=True):
    """
    Convert a CAP XML string to a GeoJSON FeatureCollection.

    :param cap_xml_string: A string containing a CAP XML document.
    :param validate: Whether to validate the CAP XML with the CAP schema for compulsory fields.
    :return: Alert data as a dictionary.
    """

    xml_dict = parse_xml_to_dict(cap_xml_string)

    alert_element_data = xml_dict.get("alert") or xml_dict.get("cap:alert")

    if not alert_element_data:
        raise CAPImportError("The loaded XML is not a valid CAP alert.")

    alert_data = {}

    for element in CAP_ALERT_ELEMENTS:
        element_name = element["name"]
        element_data = extract_element_data(element, alert_element_data, validate=validate)

        if element_data:
            alert_data[element_name] = element_data

    return alert_data



from django.conf import settings

from capeditor.pubsub.mqtt import MQTTPubSubClient
from capeditor.renderers import CapXMLRenderer
from capeditor.serializers import AlertSerializer

BROKER_URI = getattr(settings, "CAP_BROKER_URI", None)


def publish_cap_mqtt_message(cap_alert_page, topic, qos=1):
    if BROKER_URI:
        broker = {
            'url': BROKER_URI,
            'client_type': 'publisher'
        }

        # create mqtt client
        client = MQTTPubSubClient(broker)

        # serialize cap alert to xml
        data = AlertSerializer(cap_alert_page).data
        cap_xml = CapXMLRenderer().render(data)

        # publish message
        client.pub(topic, message=cap_xml, qos=qos)



# Generated by Django 4.2.11 on 2024-03-21 05:25

import django.contrib.gis.db.models.fields
from django.db import migrations, models
import django.db.models.deletion
import modelcluster.fields


class Migration(migrations.Migration):

    dependencies = [
        ('capeditor', '0003_capsetting_logo'),
    ]

    operations = [
        migrations.CreateModel(
            name='PredefinedAlertArea',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sort_order', models.IntegerField(blank=True, editable=False, null=True)),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('geom', django.contrib.gis.db.models.fields.MultiPolygonField(srid=4326, verbose_name='Area')),
                ('setting', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.PROTECT, related_name='predefined_alert_areas', to='capeditor.capsetting')),
            ],
            options={
                'verbose_name': 'Predefined Area',
                'verbose_name_plural': 'Predefined Areas',
            },
        ),
    ]



# Generated by Django 4.2.11 on 2024-06-03 11:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('capeditor', '0006_capsetting_wmo_oid'),
    ]

    operations = [
        migrations.AlterField(
            model_name='capsetting',
            name='wmo_oid',
            field=models.CharField(blank=True, help_text='WMO Register of Alerting Authorities Object Identifier (OID) of the sending institution. This will be used to generate CAP messages identifiers', max_length=255, null=True, verbose_name='WMO Register of Alerting Authorities OID'),
        ),
        migrations.AlterField(
            model_name='hazardeventtypes',
            name='event',
            field=models.CharField(help_text='Name of Hazard', max_length=35, unique=True, verbose_name='Hazard'),
        ),
    ]



# Generated by Django 4.2.11 on 2024-05-30 13:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('capeditor', '0005_alter_capsetting_sender'),
    ]

    operations = [
        migrations.AddField(
            model_name='capsetting',
            name='wmo_oid',
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name='WMO Register of Alerting Authorities OID'),
        ),
    ]



# Generated by Django 4.1.10 on 2023-10-13 14:25

from django.db import migrations, models
import django.db.models.deletion
import modelcluster.fields


class Migration(migrations.Migration):

    dependencies = [
        ('capeditor', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='capsetting',
            options={'verbose_name': 'CAP Settings'},
        ),
        migrations.RemoveField(
            model_name='capsetting',
            name='hazard_types',
        ),
        migrations.CreateModel(
            name='HazardEventTypes',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sort_order', models.IntegerField(blank=True, editable=False, null=True)),
                ('is_in_wmo_event_types_list', models.BooleanField(default=True, verbose_name='Select from WMO list of Hazards Event Types')),
                ('event', models.CharField(help_text='Name of Hazard', max_length=255, unique=True, verbose_name='Hazard')),
                ('icon', models.CharField(blank=True, help_text='Matching icon', max_length=255, null=True, verbose_name='Icon')),
                ('setting', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.PROTECT, related_name='hazard_event_types', to='capeditor.capsetting')),
            ],
            options={
                'ordering': ['sort_order'],
                'abstract': False,
            },
        ),
    ]
